package PureMessage::Policy::Spam::Engine;

use strict;
use base 'Exporter';

use ActiveState::SpamEngine qw(PMX_TYPE_STRING PMX_TYPE_INTEGER PMX_TYPE_HASH);
use PureMessage::Policy::Spam::Config qw(%config %attr);

our @EXPORT_OK = qw(datadir userdatadir traindatadir
		    new_engine new_hash new_asdb
		    get_config);

# A simple interface to the SpamEngine datadir(s)
sub datadir      { $config{datadir}      }
sub userdatadir  { $config{userdatadir}  }
sub traindatadir { $config{traindatadir} }

# Create a spam engine
sub new_engine {
    my %opts = @_;

    # If $opts{plugins} is set, only load the specified plugins; otherwise,
    # load all plugins enabled by the config files.
    my $plugins = delete $opts{plugins};
    my $logger  = delete $opts{logger};
    my $compile = delete $opts{compile};
    $compile = 1 unless defined $compile;
    die "Unknown SpamEngine options '$_'" for keys %opts;

    my $eng = ActiveState::SpamEngine->new($logger);
    $eng->add_attribute('core.datadir', PMX_TYPE_STRING, $config{datadir});
    $eng->add_attribute('core.userdatadir', PMX_TYPE_STRING, $config{userdatadir});
    $eng->add_attribute('core.plugin-dir', PMX_TYPE_STRING, $config{plugindir});
    $eng->add_attribute('core.max-bytes-scanned', PMX_TYPE_INTEGER,
			$config{max_size_checked}, 1)
	if $config{max_size_checked};
    $eng->add_attribute('plugin.sxl.enable', PMX_TYPE_INTEGER, 0)
        if $config{'disable_sxl'};
    $eng->add_attribute('core.stop-scan', PMX_TYPE_INTEGER,
			$config{stop_scan}, 1);
    $eng->add_attribute('core.local-tests-only', PMX_TYPE_INTEGER,
			$config{local_tests_only}, 1);
    $eng->add_attribute('core.profile-level', PMX_TYPE_INTEGER,
			$config{profile_level}, 1);
    $eng->load_antispam_data;

    # Set the client id to the serial number of the Spam component, if found.
    if (eval { require PureMessage::_sys }) {
	my $serial = {
	    PureMessage::_sys::lic_feature_kv({
		PureMessage::_sys::lic_features(PureMessage::_sys::lic_file())
	    }->{"PureMessage::Policy::Spam"})
	}->{"SerialNo"};
	$eng->add_attribute("core.client-id", PMX_TYPE_STRING, $serial, 1)
	    if defined $serial;
    }

    # Figure out what plugins are loaded (handles internal dependencies)
    my %loaded;
    $eng->enumerate_loaded_plugins(sub { $_[0]{$_[1]}=1; 1 }, \%loaded);

    for my $plugin (keys %{$config{plugin}}) {
	next unless $loaded{$plugin};

	# Decode the options into attributes
	for my $option (keys %{$config{plugin}{$plugin}}) {
	    next unless exists $attr{$plugin}{$option};
	    next unless defined $config{plugin}{$plugin}{$option};
	    my ($type, $name, $conv) = @{$attr{$plugin}{$option}};
	    my $attr = $name =~ /\./ ? $name : "plugin.$plugin.$name";
	    my $t    = $type eq 'str' ? PMX_TYPE_STRING :
		       $type eq 'int' ? PMX_TYPE_INTEGER :
				      die "unsupported type '$type'";
	    my $val = $config{plugin}{$plugin}{$option};
	    my @v    = $conv ? &$conv($val)
			     : ref($val) eq 'ARRAY' ? @$val : $val;
	    my $repl = 1;
	    for my $v (@v) {
		$eng->add_attribute($attr, $t, $v, $repl);
		$repl = 0;
	    }
	}
    }
    $eng->compile if $compile;

    return $eng;
}

sub new_hash {
    my %opts = @_;

    my $type = delete $opts{type};
    my $ro   = delete $opts{readonly};
    die "unknown argument '$_' to new_hash" for keys %opts;

    require ActiveState::SpamEngine::Hash;
    if ($type eq "summary") {
	my $datadir = datadir();
	return ActiveState::SpamEngine::Hash->new("$datadir/db.summary",1);
    }
    elsif ($type eq "pdelta") {
	my $datadir = userdatadir();
	return ActiveState::SpamEngine::Hash->new("$datadir/db.pdelta",$ro);
    }
    elsif ($type eq "force") {
	my $datadir = userdatadir();
	return ActiveState::SpamEngine::Hash->new("$datadir/db.force",$ro);
    }
    else {
	die;
    }
}

sub new_asdb {
    my %opts = @_;

    my $type = delete $opts{type};
    my $ro   = delete $opts{readonly};
    die "unknown argument '$_' to new_asdb" for keys %opts;

    require ActiveState::DB;
    if ($type eq "summary") {
	my $datadir = datadir();
	return ActiveState::DB->new("$datadir/asdb.summary");
    }
    else {
	die;
    }
}

# Writable config interface for the manager
sub get_config {
    my %opts = @_;

    my $writable = delete $opts{writable} || 0;

    my $config = ActiveState::Config->new(
	strict => 1,
	allow_includes => 1,
	writable => $writable,
    );
    PureMessage::Policy::Spam::Config->define_core_validators($config);
    PureMessage::Policy::Spam::Config->validate($config);

    my $config_file = PureMessage::Config::expand_config_filename('spam');
    $config->parse_file($config_file);
    return $config;
}

1;
