package PureMessage::Policy::Spam::Rules;

use strict;
use PureMessage::Version qw($PREFIX);
use ActiveState::SpamEngine::Utils qw(unpack_summary pack_force unpack_force);
use ActiveState::Config::Simple qw(config_parse_file);
use PureMessage::Policy::Spam::Engine qw(new_engine new_hash userdatadir);
use PureMessage::_sys qw();

# Change this to 1 if you'd like lots of error messages dumped to
# $PREFIX/var/log/rules.log
sub DEBUG() {0} # XXX

sub log {
    my ($self, $msg) = @_;
    chomp $msg;
    print { $self->{_fh_} } "$msg\n";
}

sub new {
    my $class = shift;
    my $ro = shift;
    $ro = 0 unless defined $ro;
    my $self = bless {
	ro	=> $ro,
    }, ref($class) || $class;
    $self->reset;
    $self;
}

sub reset {
    my $self = shift;

    my $ro = $self->{ro};
    %$self = (ro => $ro);

    if (DEBUG) {
	open (my $FH, ">>", "$PREFIX/var/log/rules.log") or die;
	select((select($FH), $| = 1)[0]);
	$self->{_fh_} = $FH;
	$self->log("reset()!");
    }

    my $s = new_hash(type => 'summary', readonly => 1);
    my $f = new_hash(type => 'force',   readonly => $ro);
    my $p = new_hash(type => 'pdelta',  readonly => $ro);
    @$self{qw(s f p)} = ($s, $f, $p);

    # Enumerate all the rules ahead of time.
    $self->{s}->enumerate(sub {
	my (undef, $key, $value) = @_;
	if ($key =~ s/^m://) {
	    $self->{map}{$key}{$_} = 1 for split ' ', $value;
	}
	elsif ($key =~ s/^g://) {
	    my $info = unpack_summary($value);
	    $self->{group}{$key} = $info;
	}
	else {
	    my $info = unpack_summary($value);
	    $self->{rules}{$key} = $info;
	}
	1;
    });

    # Parse the site rules, so that when they call getrule() on a site rule,
    # it returns more information.
    my $dd = userdatadir();
    my $rules = eval { config_parse_file("$dd/re.rules") };
    if ($@) {
	warn $@;
    }
    else {
	for my $rule (keys %$rules) {
	    my $def = $rules->{$rule};
	    next unless ref($def) eq "HASH";
	    for my $key (keys %$def) {
		next unless $key eq "Part" or $key eq "Test";
		$self->{rules}{$rule}{$key} = $def->{$key};
	    }
	}
    }
}

sub refresh {
    my $self = shift;
    $self->log("refresh()") if DEBUG;
    unless ($self->{ro}) {
	system("$PREFIX/bin/pmx-spam-compile --summary-only") == 0
	    or die "can't rebuild summary: $?";

	# Mark the fact that the data needs to be recompiled when the milter
	# is restarted.
	my $lock = "$PREFIX/etc/data/antispam/lock";
	my $locktmp = $lock . PureMessage::_sys::tid();
	open(my $FLAG, "> $locktmp")
	    or die "can't open temporary lockfile $locktmp: $!";
	print $FLAG "Data modified at ", scalar(localtime), "\n";
	close $FLAG;
	rename($locktmp => $lock)
	    or die "can't commit $locktmp to $lock: $!";
    }
    $self->reset;
}

sub rulenames {
    my $self = shift;
    if (@_) {
	my @group = @_;

	# Return all rules which belong to at least one of @group
	my @array;
	for my $rule (keys %{$self->{rules}}) {
	    if (grep $self->{map}{$rule}{$_}, @group) {
		push @array, $rule;
	    }
	}
	return @array = sort @array;
    }
    return my @array = sort keys %{$self->{rules}};
}

sub groupnames {
    my $self = shift;
    return my @array = sort keys %{$self->{group}};
}

sub getgroup {
    my ($self, $group) = @_;
    my %group = $self->{group}{$group}
		    ? %{$self->{group}{$group}}
		    : return undef;
    \%group;
}

sub memberOf {
    my ($self, $rule, $group) = @_;
    return $self->{map}{$rule}{$group};
}

sub _import {
    my ($self, $rule, $def) = @_;

    # Set the ProbabilityDelta if it's provided.
    my $pdelta = delete $def->{ProbabilityDelta};
    $self->{p}->set($rule, $pdelta) if defined $pdelta;

    # Backwards-compatibility: rename old keys to new ones
    $def->{Desc} = delete $def->{Description} if exists $def->{Description};
    $def->{w}    = delete $def->{Score}       if exists $def->{Score};

    # If it's a site rule, the {Enabled} goes into re.rules. Otherwise, it
    # needs to go into db.force along with the weight.
    my $w = delete $def->{w};
    my $e = delete $def->{Enabled};
    if (!exists $self->{rules}{$rule} or $self->memberOf($rule, "SITE")) {

	# Save the weight
	if (defined $w) {
	    my $old = $self->{f}->get($rule);
	    my $force = $old ? unpack_force($old) : {};
	    $force->{w} = $w;
	    $self->{f}->set($rule => pack_force($force));
	}

	# Now check whether anything else differs:
	my $exists = $self->getrule($rule);
	if ($exists) {
	    for my $key (keys %$def) {
		if (!$exists->{$key} or $def->{$key} ne $exists->{$key}) {
		    $self->{rules}{$rule}{$key} = $def->{$key};
		    $self->{modified} = 1;
		}
	    }
	}
	else {
	    $self->{rules}{$rule} = { %$def };
	    $self->{map}{$rule}{SITE} = 1; # new rules are SITE
	    $self->{modified} = 1;
	}
    }
    elsif (defined $w or defined $e) {
	my $old = $self->{f}->get($rule);
	my $force = $old ? unpack_force($old) : {};
	$force->{Enabled} = $e if defined $e;
	$force->{w} = $w if defined $w;
	$self->{f}->set($rule => pack_force($force));
    }
}

sub getrule {
    my ($self, $rule) = @_;
    my %rule = $self->{rules}{$rule}
		    ? %{$self->{rules}{$rule}} # make a copy
		    : return undef;
    $rule{ProbabilityDelta} = $self->{p}->get($rule) || 0;
    $rule{c_overridden} = $rule{override} eq "!" ? 1 : 0;
    $rule{c_user}       = $rule{override} eq "+" ? 1 : 0;

    # Tell the manager what fields can be edited for this rule:
    if ($rule{c_user}) {
	$rule{override} = { map { $_ => 1 } qw(Enabled Desc Part Test) };
    }
    else {
	$rule{override} = { Enabled => 1 };
    }

    \%rule;
}

sub addrule {
    my ($self, $rule, $def) = @_;
    $self->log("addrule($rule)") if DEBUG;
    return if $self->{ro};
    die "Rule $rule already exists" if exists $self->{rules}{$rule};
    $self->_import($rule, $def);
}

sub editrule {
    my ($self, $rule, $def) = @_;
    $self->log("editrule($rule)") if DEBUG;
    return if $self->{ro};
    die "Rule $rule doesn't exist" unless exists $self->{rules}{$rule};
    $self->_import($rule, $def);
}

sub revertrule {
    my ($self, $rule) = @_;
    $self->log("revertrule($rule)") if DEBUG;
    return if $self->{ro};
    $self->delrule($rule);
}

sub delrule {
    my ($self, $rule) = @_;
    return if $self->{ro};
    return unless exists $self->{rules}{$rule};

    $self->log("delrule($rule)") if DEBUG;

    # Remove the rule's overrides and pdelta
    $self->{f}->del($rule);
    $self->{p}->del($rule);

    # If it's a site rule, remove it.
    if ($self->memberOf($rule, "SITE")) {
	delete $self->{rules}{$rule};
	$self->{modified} = 1;
    }
}

# The Description field is now internally called "Desc". It must be converted
# back to "Description" when being saved to the re.rules file.
my %SAVEAS = (Desc => "Description");
sub save {
    my $self = shift;
    return if $self->{ro};

    $self->log("save()") if DEBUG;

    # If the rules were in fact modified
    if ($self->{modified}) {
	my $dd = userdatadir();
	my $dest = "$dd/re.rules";
	my $temp = "$dest.$$";

	$self->log("rules modified; writing $dest") if DEBUG;

	# Read the init block comment.
	my @header;
	open (my $OR, "<", $dest) or die "can't read $dest: $!";
	while (<$OR>) {
            if (/^#/) { push @header, $_ }
            else      { last }
	}
	close $OR;

	# Write the new file.
	open (my $FH, ">", $temp) or die "can't write $temp: $!";
	print $FH @header;
	for my $rule ($self->rulenames("SITE")) {
	    my $def = $self->getrule($rule);
	    print $FH "<$rule>\n";
	    for my $k (qw(Desc Enabled Part Test)) {
		next unless exists $def->{$k};

		my $wk = $SAVEAS{$k} || $k;
		my $wv = _quote($wk, $def->{$k});
		next unless defined $wk and defined $wv;

		print $FH "    $wk = $wv\n";
	    }
	    print $FH "</$rule>\n";
	}
	close $FH or die "can't write $temp: $!";
	rename $temp, $dest or die "can't rename $temp to $dest: $!";
    }

    # Update the databases.
    $self->{f}->save;
    $self->{p}->save;
    delete @$self{qw(f p s)};

    $self->refresh;

    1;
}

sub _quote {
    my ($wk, $v) = @_;
    if ($wk eq "Description") {
	$v =~ s/([\\"])/\\$1/g;
	return qq{"$v"};
    }
    elsif ($wk eq "Enabled") {
	# If Enabled is true (the default) don't bother emitting it.
	if ($v) { return undef }
	return "no";
    }
    return $v;
}

1;

=head1 NAME

PureMessage::Policy::Spam::Rules - generic interface to AntiSpam rules

=head1 SYNOPSIS

  use PureMessage::Policy::Spam::Rules;

  my $rules = PureMessage::Policy::Spam::Rules->new;

  # Enumerate rules
  for my $rule ($rules->rulenames) {
      print "Rule $rule:\n";
      my $def = $rules->getrule($rule);
      print "  Enabled:          $def->{Enabled}\n";
      print "  Score:            $def->{w}\n";
      print "  ProbabilityDelta: $def->{ProbabilityDelta}\n";
  }

  # Modifying and saving rules
  my $change = $rules->getrule("BAYES_00_10");
  $change->{w} = 0;
  $change->{ProbabilityDelta} = -5;
  $rules->editrule("BAYES_00_10", $change);

  $change = $rules->getrule("ADDR_FREE");
  $change->{Test} = "/abcdefg/i";
  $rules->editrule("ADDR_FREE", $change);

=head1 DESCRIPTION

PureMessage::Policy::Spam::Rules provides a generic interface to display and
modify C<rules> used to detect spam. The interface hides details like whether
the rule is ActiveState's or the user's, and provides a simple way to modify
attributes of a rule.

=head2 Rule Attributes

Rules are represented as a hash. Several keys are always defined, and are
important:

=over

=item Desc

A description (in English) of what the rule identifies in the message.

=item LongDesc

A description (in English) of what the rule identifies, plus a bit more
information about *how* it does it. Designed to be used in the PureMessage
Manager.

=item Enabled

A boolean (C<0> or C<1>) that specifies whether this rule is enabled.

=item w

The score for this rule. This is always editable.

=item ProbabilityDelta

The probability delta for this rule. This is always editable. Should be a
number between 0 and 100 (it is divided by 100 internally).

=item c_user

A boolean (C<0> or C<1>) that specifies whether the rule is a
customer-specific rule, or at least one attribute has been overridden by a
customer-specific change. Currently there is no way to see which attribute has
been overridden.

=item c_overridden

A boolean (C<0> or C<1>) that specifies whether the rule belongs to the default
rules set (ActiveState) and has been modified by a customer-specific change. 
This includes modifications made to the Score, ProbabilityDelta and attributes 
from C<override> list.

=back

The C<SITE> feature group uses the following attributes:

=over

=item Part

The part of the message being tested.

=item Test

A regular expression to be run on the part.

=back

=head2 API

Use the following methods to view, edit, add, and delete rules.

=over

=item rulenames()

   @rules = $rules->rulenames;

Returns a sorted list of rule names defined in the system.

=item rulenames($GROUP)

   @rules = $rules->rulenames("SITE");

Returns a sorted list of rule names defined in the specified group.

=item rulenames($GROUP1, $GROUP2, ...)

   @rules = $rules->rulenames("SITE", "HEUR");

Returns a sorted list of rule names belonging to any of the specified groups.

=item getrule()

   $def = $rules->getrule($name);

Returns the attribute hash describing the named rule, or C<undef> if the rules
does not exist.

For example, this is the definition of C<ACCOUNT_CLICK>:

   {
       Desc             => "Click to perform an action on an account",
       Enabled          => 1,
       LongDesc         => "...",
       ProbabilityDelta => 0,
       w                => 0.137,
   }

=item editrule()

   $rules->editrule($name, $def);

Schedules the modified rule definition for saving. Only the weights database,
probability database, and user-specific configuration files are ever modified.
Only those keys which are overridable are inspected for changes; other fields
are ignored.

=item addrule()

   $rules->addrule($name, $def);

Similar to C<editrule()>, this adds a user-specific rule. It throws an
exception if the rule already exists (use editrule() or delrule()).

=item delrule()

   $rules->delrule($name, $def);

Deletes a user-specified rule. If the rule was not a custom or overridden
rule, this will have no effect, since the ActiveState rules are never
overwritten.

=item revertrule()

   $rules->revertrule($name, $def);

Reverts an overridden rule. It's similar to C<delrule()> but it also deletes 
user-specific weights and pdeltas modifications of the default rules.
If the rule was not a custom or overridden rule, this will have no effect.

=item save()

   $rules->save;

Saves changes to the rules, scores and probability delta files.

=back

=cut
